@@ -36,10 +36,17 @@
 #include <proto/log.h>
 #include <proto/freq_ctr.h>
 
+
+__decl_hathreads(extern HA_SPINLOCK_T idle_conn_srv_lock);
+extern struct eb_root idle_conn_srv;
+extern struct task *idle_conn_task;
+extern struct task *idle_conn_cleanup[MAX_THREADS];
+extern struct list toremove_connections[MAX_THREADS];
+
 int srv_downtime(const struct server *s);
 int srv_lastsession(const struct server *s);
 int srv_getinter(const struct check *check);
-int parse_server(const char *file, int linenum, char **args, struct proxy *curproxy, struct proxy *defproxy);
+int parse_server(const char *file, int linenum, char **args, struct proxy *curproxy, struct proxy *defproxy, int parse_addr);
 int update_server_addr(struct server *s, void *ip, int ip_sin_family, const char *updater);
 const char *update_server_addr_port(struct server *s, const char *addr, const char *port, char *updater);
 struct server *server_find_by_id(struct proxy *bk, int id);
@@ -59,17 +66,18 @@ int snr_resolution_cb(struct dns_requester *requester, struct dns_nameserver *na
 int snr_resolution_error_cb(struct dns_requester *requester, int error_code);
 struct server *snr_check_ip_callback(struct server *srv, void *ip, unsigned char *ip_family);
 struct task *srv_cleanup_idle_connections(struct task *task, void *ctx, unsigned short state);
+struct task *srv_cleanup_toremove_connections(struct task *task, void *context, unsigned short state);
 
 /* increase the number of cumulated connections on the designated server */
-static void inline srv_inc_sess_ctr(struct server *s)
+static inline void srv_inc_sess_ctr(struct server *s)
 {
-	HA_ATOMIC_ADD(&s->counters.cum_sess, 1);
+	_HA_ATOMIC_ADD(&s->counters.cum_sess, 1);
 	HA_ATOMIC_UPDATE_MAX(&s->counters.sps_max,
 			     update_freq_ctr(&s->sess_per_sec, 1));
 }
 
 /* set the time of last session on the designated server */
-static void inline srv_set_sess_last(struct server *s)
+static inline void srv_set_sess_last(struct server *s)
 {
 	s->counters.last_sess = now.tv_sec;
 }
@@ -243,23 +251,36 @@ static inline int srv_add_to_idle_list(struct server *srv, struct connection *co
 	    (srv->max_idle_conns == -1 || srv->max_idle_conns > srv->curr_idle_conns) &&
 	    !(conn->flags & CO_FL_PRIVATE) &&
 	    ((srv->proxy->options & PR_O_REUSE_MASK) != PR_O_REUSE_NEVR) &&
-	    !conn->mux->used_streams(conn) && conn->mux->avail_streams(conn)) {
+	    !conn->mux->used_streams(conn) && conn->mux->avail_streams(conn) &&
+	    ha_used_fds < global.tune.pool_low_count) {
 		int retadd;
 
-		retadd = HA_ATOMIC_ADD(&srv->curr_idle_conns, 1);
-		if (retadd >= srv->max_idle_conns) {
-			HA_ATOMIC_SUB(&srv->curr_idle_conns, 1);
+		retadd = _HA_ATOMIC_ADD(&srv->curr_idle_conns, 1);
+		if (retadd > srv->max_idle_conns) {
+			_HA_ATOMIC_SUB(&srv->curr_idle_conns, 1);
 			return 0;
 		}
 		LIST_DEL(&conn->list);
-		LIST_ADDQ(&srv->idle_orphan_conns[tid], &conn->list);
+		LIST_ADDQ_LOCKED(&srv->idle_orphan_conns[tid], &conn->list);
 		srv->curr_idle_thr[tid]++;
 
 		conn->idle_time = now_ms;
-		if (!(task_in_wq(srv->idle_task[tid])) &&
-		    !(task_in_rq(srv->idle_task[tid])))
-			task_schedule(srv->idle_task[tid],
-			    tick_add(now_ms, srv->pool_purge_delay));
+		__ha_barrier_full();
+		if ((volatile void *)srv->idle_node.node.leaf_p == NULL) {
+			HA_SPIN_LOCK(OTHER_LOCK, &idle_conn_srv_lock);
+			if ((volatile void *)srv->idle_node.node.leaf_p == NULL) {
+				srv->idle_node.key = tick_add(srv->pool_purge_delay,
+				                              now_ms);
+				eb32_insert(&idle_conn_srv, &srv->idle_node);
+				if (!task_in_wq(idle_conn_task) && !
+				    task_in_rq(idle_conn_task)) {
+					task_schedule(idle_conn_task,
+					              srv->idle_node.key);
+				}
+
+			}
+			HA_SPIN_UNLOCK(OTHER_LOCK, &idle_conn_srv_lock);
+		}
 		return 1;
 	}
 	return 0;
