@@ -600,7 +600,7 @@ unsigned int http_get_htx_fhdr(const struct htx *htx, const struct ist hdr,
 	return 1;
 }
 
-static struct htx *http_str_to_htx(struct buffer *buf, struct ist raw)
+int http_str_to_htx(struct buffer *buf, struct ist raw)
 {
 	struct htx *htx;
 	struct htx_sl *sl;
@@ -610,11 +610,17 @@ static struct htx *http_str_to_htx(struct buffer *buf, struct ist raw)
 	unsigned int flags = HTX_SL_F_IS_RESP;
 	int ret = 0;
 
+	b_reset(buf);
+	if (!raw.len) {
+		buf->size = 0;
+		buf->area = malloc(raw.len);
+		return 1;
+	}
+
 	buf->size = global.tune.bufsize;
 	buf->area = (char *)malloc(buf->size);
 	if (!buf->area)
 		goto error;
-	b_reset(buf);
 
 	h1m_init_res(&h1m);
 	h1m.flags |= H1_MF_NO_PHDR;
@@ -663,12 +669,12 @@ static struct htx *http_str_to_htx(struct buffer *buf, struct ist raw)
 	if (!htx_add_endof(htx, HTX_BLK_EOM))
 		goto error;
 
-	return htx;
+	return 1;
 
 error:
 	if (buf->size)
 		free(buf->area);
-	return NULL;
+	return 0;
 }
 
 static int http_htx_init(void)
@@ -680,7 +686,7 @@ static int http_htx_init(void)
 	int err_code = 0;
 
 	for (px = proxies_list; px; px = px->next) {
-		if (px->mode != PR_MODE_HTTP || !(px->options2 & PR_O2_USE_HTX))
+		if (!(px->options2 & PR_O2_USE_HTX))
 			continue;
 
 		for (rc = 0; rc < HTTP_ERR_SIZE; rc++) {
