@@ -39,11 +39,9 @@
 #include <proto/frontend.h>
 #include <proto/log.h>
 #include <proto/sample.h>
+#include <proto/ssl_sock.h>
 #include <proto/stream.h>
 #include <proto/stream_interface.h>
-#ifdef USE_OPENSSL
-#include <proto/ssl_sock.h>
-#endif
 
 struct log_fmt {
 	char *name;
@@ -84,23 +82,18 @@ static const struct log_fmt log_formats[LOG_FORMATS] = {
 	},
 };
 
-#define FD_SETS_ARE_BITFIELDS
-#ifdef FD_SETS_ARE_BITFIELDS
 /*
  * This map is used with all the FD_* macros to check whether a particular bit
- * is set or not. Each bit represents an ACSII code. FD_SET() sets those bytes
- * which should be escaped. When FD_ISSET() returns non-zero, it means that the
- * byte should be escaped. Be careful to always pass bytes from 0 to 255
- * exclusively to the macros.
+ * is set or not. Each bit represents an ACSII code. ha_bit_set() sets those
+ * bytes which should be escaped. When ha_bit_test() returns non-zero, it means
+ * that the byte should be escaped. Be careful to always pass bytes from 0 to
+ * 255 exclusively to the macros.
  */
-fd_set rfc5424_escape_map[(sizeof(fd_set) > (256/8)) ? 1 : ((256/8) / sizeof(fd_set))];
-fd_set hdr_encode_map[(sizeof(fd_set) > (256/8)) ? 1 : ((256/8) / sizeof(fd_set))];
-fd_set url_encode_map[(sizeof(fd_set) > (256/8)) ? 1 : ((256/8) / sizeof(fd_set))];
-fd_set http_encode_map[(sizeof(fd_set) > (256/8)) ? 1 : ((256/8) / sizeof(fd_set))];
+long rfc5424_escape_map[(256/8) / sizeof(long)];
+long hdr_encode_map[(256/8) / sizeof(long)];
+long url_encode_map[(256/8) / sizeof(long)];
+long http_encode_map[(256/8) / sizeof(long)];
 
-#else
-#error "Check if your OS uses bitfields for fd_sets"
-#endif
 
 const char *log_facilities[NB_LOG_FACILITIES] = {
 	"kern", "user", "mail", "daemon",
@@ -232,11 +225,13 @@ unsigned int dropped_logs = 0;
  * update_log_hdr().
  */
 THREAD_LOCAL char *logheader = NULL;
+THREAD_LOCAL char *logheader_end = NULL;
 
 /* This is a global syslog header for messages in RFC5424 format. It is
  * updated by update_log_hdr_rfc5424().
  */
 THREAD_LOCAL char *logheader_rfc5424 = NULL;
+THREAD_LOCAL char *logheader_rfc5424_end = NULL;
 
 /* This is a global syslog message buffer, common to all outgoing
  * messages. It contains only the data part.
@@ -377,7 +372,7 @@ int parse_logformat_var_args(char *args, struct logformat_node *node, char **err
 int parse_logformat_var(char *arg, int arg_len, char *var, int var_len, struct proxy *curproxy, struct list *list_format, int *defoptions, char **err)
 {
 	int j;
-	struct logformat_node *node;
+	struct logformat_node *node = NULL;
 
 	for (j = 0; logformat_keywords[j].name; j++) { // search a log type
 		if (strlen(logformat_keywords[j].name) == var_len &&
@@ -386,14 +381,14 @@ int parse_logformat_var(char *arg, int arg_len, char *var, int var_len, struct p
 				node = calloc(1, sizeof(*node));
 				if (!node) {
 					memprintf(err, "out of memory error");
-					return 0;
+					goto error_free;
 				}
 				node->type = logformat_keywords[j].type;
 				node->options = *defoptions;
 				if (arg_len) {
 					node->arg = my_strndup(arg, arg_len);
 					if (!parse_logformat_var_args(node->arg, node, err))
-						return 0;
+						goto error_free;
 				}
 				if (node->type == LOG_FMT_GLOBAL) {
 					*defoptions = node->options;
@@ -402,7 +397,7 @@ int parse_logformat_var(char *arg, int arg_len, char *var, int var_len, struct p
 				} else {
 					if (logformat_keywords[j].config_callback &&
 					    logformat_keywords[j].config_callback(node, curproxy) != 0) {
-						return 0;
+						goto error_free;
 					}
 					curproxy->to_log |= logformat_keywords[j].lw;
 					LIST_ADDQ(list_format, &node->list);
@@ -415,7 +410,7 @@ int parse_logformat_var(char *arg, int arg_len, char *var, int var_len, struct p
 			} else {
 				memprintf(err, "format variable '%s' is reserved for HTTP mode",
 				          logformat_keywords[j].name);
-				return 0;
+				goto error_free;
 			}
 		}
 	}
@@ -424,6 +419,12 @@ int parse_logformat_var(char *arg, int arg_len, char *var, int var_len, struct p
 	var[var_len] = 0;
 	memprintf(err, "no such format variable '%s'. If you wanted to emit the '%%' character verbatim, you need to use '%%%%'", var);
 	var[var_len] = j;
+
+  error_free:
+	if (node) {
+		free(node->arg);
+		free(node);
+	}
 	return 0;
 }
 
@@ -476,8 +477,8 @@ int add_to_logformat_list(char *start, char *end, int type, struct list *list_fo
 int add_sample_to_logformat_list(char *text, char *arg, int arg_len, struct proxy *curpx, struct list *list_format, int options, int cap, char **err)
 {
 	char *cmd[2];
-	struct sample_expr *expr;
-	struct logformat_node *node;
+	struct sample_expr *expr = NULL;
+	struct logformat_node *node = NULL;
 	int cmd_arg;
 
 	cmd[0] = text;
@@ -487,13 +488,13 @@ int add_sample_to_logformat_list(char *text, char *arg, int arg_len, struct prox
 	expr = sample_parse_expr(cmd, &cmd_arg, curpx->conf.args.file, curpx->conf.args.line, err, &curpx->conf.args);
 	if (!expr) {
 		memprintf(err, "failed to parse sample expression <%s> : %s", text, *err);
-		return 0;
+		goto error_free;
 	}
 
 	node = calloc(1, sizeof(*node));
 	if (!node) {
 		memprintf(err, "out of memory error");
-		return 0;
+		goto error_free;
 	}
 	node->type = LOG_FMT_EXPR;
 	node->expr = expr;
@@ -502,7 +503,7 @@ int add_sample_to_logformat_list(char *text, char *arg, int arg_len, struct prox
 	if (arg_len) {
 		node->arg = my_strndup(arg, arg_len);
 		if (!parse_logformat_var_args(node->arg, node, err))
-			return 0;
+			goto error_free;
 	}
 	if (expr->fetch->val & cap & SMP_VAL_REQUEST)
 		node->options |= LOG_OPT_REQ_CAP; /* fetch method is request-compatible */
@@ -511,11 +512,9 @@ int add_sample_to_logformat_list(char *text, char *arg, int arg_len, struct prox
 		node->options |= LOG_OPT_RES_CAP; /* fetch method is response-compatible */
 
 	if (!(expr->fetch->val & cap)) {
-		free(node);
-		node = NULL;
 		memprintf(err, "sample fetch <%s> may not be reliably used here because it needs '%s' which is not available here",
 		          text, sample_src_names(expr->fetch->use));
-		return 0;
+		goto error_free;
 	}
 
 	/* check if we need to allocate an hdr_idx struct for HTTP parsing */
@@ -530,6 +529,14 @@ int add_sample_to_logformat_list(char *text, char *arg, int arg_len, struct prox
 	curpx->to_log |= LW_REQ;
 	LIST_ADDQ(list_format, &node->list);
 	return 1;
+
+  error_free:
+	release_sample_expr(expr);
+	if (node) {
+		free(node->arg);
+		free(node);
+	}
+	return 0;
 }
 
 /*
@@ -567,6 +574,8 @@ int parse_logformat_string(const char *fmt, struct proxy *curproxy, struct list
 	/* flush the list first. */
 	list_for_each_entry_safe(tmplf, back, list_format, list) {
 		LIST_DEL(&tmplf->list);
+		release_sample_expr(tmplf->expr);
+		free(tmplf->arg);
 		free(tmplf);
 	}
 
@@ -691,6 +700,79 @@ int parse_logformat_string(const char *fmt, struct proxy *curproxy, struct list
 	return 1;
 }
 
+/*
+ * Parse the first range of indexes from a string made of a list of comma seperated
+ * ranges of indexes. Note that an index may be considered as a particular range
+ * with a high limit to the low limit.
+ */
+int get_logsrv_smp_range(unsigned int *low, unsigned int *high, char **arg, char **err)
+{
+	char *end, *p;
+
+	*low = *high = 0;
+
+	p = *arg;
+	end = strchr(p, ',');
+	if (!end)
+		end = p + strlen(p);
+
+	*high = *low = read_uint((const char **)&p, end);
+	if (!*low || (p != end && *p != '-'))
+		goto err;
+
+	if (p == end)
+		goto done;
+
+	p++;
+	*high = read_uint((const char **)&p, end);
+	if (!*high || *high <= *low || p != end)
+		goto err;
+
+ done:
+	if (*end == ',')
+		end++;
+	*arg = end;
+	return 1;
+
+ err:
+	memprintf(err, "wrong sample range '%s'", *arg);
+	return 0;
+}
+
+/*
+ * Returns 1 if the range defined by <low> and <high> overlaps
+ * one of them in <rgs> array of ranges with <sz> the size of this
+ * array, 0 if not.
+ */
+int smp_log_ranges_overlap(struct smp_log_range *rgs, size_t sz,
+                           unsigned int low, unsigned int high, char **err)
+{
+	size_t i;
+
+	for (i = 0; i < sz; i++) {
+		if ((low  >= rgs[i].low && low  <= rgs[i].high) ||
+		    (high >= rgs[i].low && high <= rgs[i].high)) {
+			memprintf(err, "ranges are overlapping");
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
+int smp_log_range_cmp(const void *a, const void *b)
+{
+	const struct smp_log_range *rg_a = a;
+	const struct smp_log_range *rg_b = b;
+
+	if (rg_a->high < rg_b->low)
+		return -1;
+	else if (rg_a->low > rg_b->high)
+		return 1;
+
+	return 0;
+}
+
 /*
  * Parse "log" keyword and update <logsrvs> list accordingly.
  *
@@ -808,6 +890,77 @@ int parse_logsrv(char **args, struct list *logsrvs, int do_del, char **err)
 		cur_arg += 2;
 	}
 
+	if (strcmp(args[cur_arg], "sample") == 0) {
+		unsigned low, high;
+		char *p, *beg, *end, *smp_sz_str;
+		struct smp_log_range *smp_rgs = NULL;
+		size_t smp_rgs_sz = 0, smp_sz = 0, new_smp_sz;
+
+		p = args[cur_arg+1];
+		smp_sz_str = strchr(p, ':');
+		if (!smp_sz_str) {
+			memprintf(err, "Missing sample size");
+			goto error;
+		}
+
+		*smp_sz_str++ = '\0';
+
+		end = p + strlen(p);
+
+		while (p != end) {
+			if (!get_logsrv_smp_range(&low, &high, &p, err))
+				goto error;
+
+			if (smp_rgs && smp_log_ranges_overlap(smp_rgs, smp_rgs_sz, low, high, err))
+				goto error;
+
+			smp_rgs = my_realloc2(smp_rgs, (smp_rgs_sz + 1) * sizeof *smp_rgs);
+			if (!smp_rgs) {
+				memprintf(err, "out of memory error");
+				goto error;
+			}
+
+			smp_rgs[smp_rgs_sz].low = low;
+			smp_rgs[smp_rgs_sz].high = high;
+			smp_rgs[smp_rgs_sz].sz = high - low + 1;
+			smp_rgs[smp_rgs_sz].curr_idx = 0;
+			if (smp_rgs[smp_rgs_sz].high > smp_sz)
+				smp_sz = smp_rgs[smp_rgs_sz].high;
+			smp_rgs_sz++;
+		}
+
+		if (smp_rgs == NULL) {
+			memprintf(err, "no sampling ranges given");
+			goto error;
+		}
+
+		beg = smp_sz_str;
+		end = beg + strlen(beg);
+		new_smp_sz = read_uint((const char **)&beg, end);
+		if (!new_smp_sz || beg != end) {
+			memprintf(err, "wrong sample size '%s' for sample range '%s'",
+						   smp_sz_str, args[cur_arg+1]);
+			goto error;
+		}
+
+		if (new_smp_sz < smp_sz) {
+			memprintf(err, "sample size %zu should be greater or equal to "
+						   "%zu the maximum of the high ranges limits",
+						   new_smp_sz, smp_sz);
+			goto error;
+		}
+		smp_sz = new_smp_sz;
+
+		/* Let's order <smp_rgs> array. */
+		qsort(smp_rgs, smp_rgs_sz, sizeof(struct smp_log_range), smp_log_range_cmp);
+
+		logsrv->lb.smp_rgs = smp_rgs;
+		logsrv->lb.smp_rgs_sz = smp_rgs_sz;
+		logsrv->lb.smp_sz = smp_sz;
+
+		cur_arg += 2;
+	}
+	HA_SPIN_INIT(&logsrv->lock);
 	/* parse the facility */
 	logsrv->facility = get_log_facility(args[cur_arg]);
 	if (logsrv->facility < 0) {
@@ -1000,7 +1153,7 @@ int get_log_facility(const char *fac)
  * <escape>.
  */
 static char *lf_encode_string(char *start, char *stop,
-                              const char escape, const fd_set *map,
+                              const char escape, const long *map,
                               const char *string,
                               struct logformat_node *node)
 {
@@ -1008,8 +1161,8 @@ static char *lf_encode_string(char *start, char *stop,
 		if (start < stop) {
 			stop--; /* reserve one byte for the final '\0' */
 			while (start < stop && *string != '\0') {
-				if (!FD_ISSET((unsigned char)(*string), map)) {
-					if (!FD_ISSET((unsigned char)(*string), rfc5424_escape_map))
+				if (!ha_bit_test((unsigned char)(*string), map)) {
+					if (!ha_bit_test((unsigned char)(*string), rfc5424_escape_map))
 						*start++ = *string;
 					else {
 						if (start + 2 >= stop)
@@ -1045,7 +1198,7 @@ static char *lf_encode_string(char *start, char *stop,
  * <escape>.
  */
 static char *lf_encode_chunk(char *start, char *stop,
-                             const char escape, const fd_set *map,
+                             const char escape, const long *map,
                              const struct buffer *chunk,
                              struct logformat_node *node)
 {
@@ -1058,8 +1211,8 @@ static char *lf_encode_chunk(char *start, char *stop,
 
 			stop--; /* reserve one byte for the final '\0' */
 			while (start < stop && str < end) {
-				if (!FD_ISSET((unsigned char)(*str), map)) {
-					if (!FD_ISSET((unsigned char)(*str), rfc5424_escape_map))
+				if (!ha_bit_test((unsigned char)(*str), map)) {
+					if (!ha_bit_test((unsigned char)(*str), rfc5424_escape_map))
 						*start++ = *str;
 					else {
 						if (start + 2 >= stop)
@@ -1210,11 +1363,10 @@ char *lf_port(char *dst, const struct sockaddr *sockaddr, size_t size, const str
 static char *update_log_hdr(const time_t time)
 {
 	static THREAD_LOCAL long tvsec;
-	static THREAD_LOCAL char *dataptr = NULL; /* backup of last end of header, NULL first time */
 	static THREAD_LOCAL struct buffer host = { };
 	static THREAD_LOCAL int sep = 0;
 
-	if (unlikely(time != tvsec || dataptr == NULL)) {
+	if (unlikely(time != tvsec || logheader_end == NULL)) {
 		/* this string is rebuild only once a second */
 		struct tm tm;
 		int hdr_len;
@@ -1240,12 +1392,12 @@ static char *update_log_hdr(const time_t time)
 		if (hdr_len < 0 || hdr_len > global.max_syslog_len)
 			hdr_len = global.max_syslog_len;
 
-		dataptr = logheader + hdr_len;
+		logheader_end = logheader + hdr_len;
 	}
 
-	dataptr[0] = 0; // ensure we get rid of any previous attempt
+	logheader_end[0] = 0; // ensure we get rid of any previous attempt
 
-	return dataptr;
+	return logheader_end;
 }
 
 /* Re-generate time-based part of the syslog header in RFC5424 format at
@@ -1255,10 +1407,9 @@ static char *update_log_hdr(const time_t time)
 static char *update_log_hdr_rfc5424(const time_t time)
 {
 	static THREAD_LOCAL long tvsec;
-	static THREAD_LOCAL char *dataptr = NULL; /* backup of last end of header, NULL first time */
 	const char *gmt_offset;
 
-	if (unlikely(time != tvsec || dataptr == NULL)) {
+	if (unlikely(time != tvsec || logheader_rfc5424_end == NULL)) {
 		/* this string is rebuild only once a second */
 		struct tm tm;
 		int hdr_len;
@@ -1280,12 +1431,12 @@ static char *update_log_hdr_rfc5424(const time_t time)
 		if (hdr_len < 0 || hdr_len > global.max_syslog_len)
 			hdr_len = global.max_syslog_len;
 
-		dataptr = logheader_rfc5424 + hdr_len;
+		logheader_rfc5424_end = logheader_rfc5424 + hdr_len;
 	}
 
-	dataptr[0] = 0; // ensure we get rid of any previous attempt
+	logheader_rfc5424_end[0] = 0; // ensure we get rid of any previous attempt
 
-	return dataptr;
+	return logheader_rfc5424_end;
 }
 
 /*
@@ -1310,13 +1461,16 @@ void send_log(struct proxy *p, int level, const char *format, ...)
 }
 
 /*
- * This function sends a syslog message.
- * It doesn't care about errors nor does it report them.
+ * This function sends a syslog message to <logsrv>.
+ * <pid_str> is the string to be used for the PID of the caller, <pid_size> is length.
+ * Same thing for <sd> and <sd_size> which are used for the structured-data part
+ * in RFC5424 formatted syslog messages, and <tag_str> and <tag_size> the syslog tag.
  * It overrides the last byte of the message vector with an LF character.
- * The arguments <sd> and <sd_size> are used for the structured-data part
- * in RFC5424 formatted syslog messages.
+ * Does not return any error,
  */
-void __send_log(struct proxy *p, int level, char *message, size_t size, char *sd, size_t sd_size)
+static inline void __do_send_log(struct logsrv *logsrv, int nblogger, char *pid_str, size_t pid_size,
+                                 int level, char *message, size_t size, char *sd, size_t sd_size,
+                                 char *tag_str, size_t tag_size)
 {
 	static THREAD_LOCAL struct iovec iovec[NB_MSG_IOVEC_ELEMENTS] = { };
 	static THREAD_LOCAL struct msghdr msghdr = {
@@ -1326,22 +1480,244 @@ void __send_log(struct proxy *p, int level, char *message, size_t size, char *sd
 	static THREAD_LOCAL int logfdunix = -1;	/* syslog to AF_UNIX socket */
 	static THREAD_LOCAL int logfdinet = -1;	/* syslog to AF_INET socket */
 	static THREAD_LOCAL char *dataptr = NULL;
-	int fac_level;
-	struct list *logsrvs = NULL;
-	struct logsrv *tmp = NULL;
-	int nblogger;
+	time_t time = date.tv_sec;
 	char *hdr, *hdr_ptr;
 	size_t hdr_size;
-	time_t time = date.tv_sec;
-	struct buffer *tag = &global.log_tag;
-	static THREAD_LOCAL int curr_pid;
-	static THREAD_LOCAL char pidstr[100];
-	static THREAD_LOCAL struct buffer pid;
+	int fac_level;
+	int *plogfd;
+	char *pid_sep1 = "", *pid_sep2 = "";
+	char logheader_short[3];
+	int sent;
+	int maxlen;
+	int hdr_max = 0;
+	int tag_max = 0;
+	int pid_sep1_max = 0;
+	int pid_max = 0;
+	int pid_sep2_max = 0;
+	int sd_max = 0;
+	int max = 0;
 
 	msghdr.msg_iov = iovec;
 
 	dataptr = message;
 
+	if (logsrv->addr.ss_family == AF_UNSPEC) {
+		/* the socket's address is a file descriptor */
+		plogfd = (int *)&((struct sockaddr_in *)&logsrv->addr)->sin_addr.s_addr;
+		if (unlikely(!((struct sockaddr_in *)&logsrv->addr)->sin_port)) {
+			/* FD not yet initialized to non-blocking mode.
+			 * DON'T DO IT ON A TERMINAL!
+			 */
+			if (!isatty(*plogfd))
+				fcntl(*plogfd, F_SETFL, O_NONBLOCK);
+			((struct sockaddr_in *)&logsrv->addr)->sin_port = 1;
+		}
+	}
+	else if (logsrv->addr.ss_family == AF_UNIX)
+		plogfd = &logfdunix;
+	else
+		plogfd = &logfdinet;
+
+	if (unlikely(*plogfd < 0)) {
+		/* socket not successfully initialized yet */
+		if ((*plogfd = socket(logsrv->addr.ss_family, SOCK_DGRAM,
+							  (logsrv->addr.ss_family == AF_UNIX) ? 0 : IPPROTO_UDP)) < 0) {
+			static char once;
+
+			if (!once) {
+				once = 1; /* note: no need for atomic ops here */
+				ha_alert("socket() failed in logger #%d: %s (errno=%d)\n",
+						 nblogger, strerror(errno), errno);
+			}
+			return;
+		} else {
+			/* we don't want to receive anything on this socket */
+			setsockopt(*plogfd, SOL_SOCKET, SO_RCVBUF, &zero, sizeof(zero));
+			/* does nothing under Linux, maybe needed for others */
+			shutdown(*plogfd, SHUT_RD);
+			fcntl(*plogfd, F_SETFD, fcntl(*plogfd, F_GETFD, FD_CLOEXEC) | FD_CLOEXEC);
+		}
+	}
+
+	switch (logsrv->format) {
+	case LOG_FORMAT_RFC3164:
+		hdr = logheader;
+		hdr_ptr = update_log_hdr(time);
+		break;
+
+	case LOG_FORMAT_RFC5424:
+		hdr = logheader_rfc5424;
+		hdr_ptr = update_log_hdr_rfc5424(time);
+		sd_max = sd_size; /* the SD part allowed only in RFC5424 */
+		break;
+
+	case LOG_FORMAT_SHORT:
+		/* all fields are known, skip the header generation */
+		hdr = logheader_short;
+		hdr[0] = '<';
+		hdr[1] = '0' + MAX(level, logsrv->minlvl);
+		hdr[2] = '>';
+		hdr_ptr = hdr;
+		hdr_max = 3;
+		maxlen = logsrv->maxlen - hdr_max;
+		max = MIN(size, maxlen) - 1;
+		goto send;
+
+	case LOG_FORMAT_RAW:
+		/* all fields are known, skip the header generation */
+		hdr_ptr = hdr = "";
+		hdr_max = 0;
+		maxlen = logsrv->maxlen;
+		max = MIN(size, maxlen) - 1;
+		goto send;
+
+	default:
+		return; /* must never happen */
+	}
+
+	hdr_size = hdr_ptr - hdr;
+
+	/* For each target, we may have a different facility.
+	 * We can also have a different log level for each message.
+	 * This induces variations in the message header length.
+	 * Since we don't want to recompute it each time, nor copy it every
+	 * time, we only change the facility in the pre-computed header,
+	 * and we change the pointer to the header accordingly.
+	 */
+	fac_level = (logsrv->facility << 3) + MAX(level, logsrv->minlvl);
+	hdr_ptr = hdr + 3; /* last digit of the log level */
+	do {
+		*hdr_ptr = '0' + fac_level % 10;
+		fac_level /= 10;
+		hdr_ptr--;
+	} while (fac_level && hdr_ptr > hdr);
+	*hdr_ptr = '<';
+
+	hdr_max = hdr_size - (hdr_ptr - hdr);
+
+	/* time-based header */
+	if (unlikely(hdr_size >= logsrv->maxlen)) {
+		hdr_max = MIN(hdr_max, logsrv->maxlen) - 1;
+		sd_max = 0;
+		goto send;
+	}
+
+	maxlen = logsrv->maxlen - hdr_max;
+
+	/* tag */
+	tag_max = tag_size;
+	if (unlikely(tag_max >= maxlen)) {
+		tag_max = maxlen - 1;
+		sd_max = 0;
+		goto send;
+	}
+
+	maxlen -= tag_max;
+
+	/* first pid separator */
+	pid_sep1_max = log_formats[logsrv->format].pid.sep1.data;
+	if (unlikely(pid_sep1_max >= maxlen)) {
+		pid_sep1_max = maxlen - 1;
+		sd_max = 0;
+		goto send;
+	}
+
+	pid_sep1 = log_formats[logsrv->format].pid.sep1.area;
+	maxlen -= pid_sep1_max;
+
+	/* pid */
+	pid_max = pid_size;
+	if (unlikely(pid_size >= maxlen)) {
+		pid_size = maxlen - 1;
+		sd_max = 0;
+		goto send;
+	}
+
+	maxlen -= pid_size;
+
+	/* second pid separator */
+	pid_sep2_max = log_formats[logsrv->format].pid.sep2.data;
+	if (unlikely(pid_sep2_max >= maxlen)) {
+		pid_sep2_max = maxlen - 1;
+		sd_max = 0;
+		goto send;
+	}
+
+	pid_sep2 = log_formats[logsrv->format].pid.sep2.area;
+	maxlen -= pid_sep2_max;
+
+	/* structured-data */
+	if (sd_max >= maxlen) {
+		sd_max = maxlen - 1;
+		goto send;
+	}
+
+	max = MIN(size, maxlen - sd_max) - 1;
+send:
+	iovec[0].iov_base = hdr_ptr;
+	iovec[0].iov_len  = hdr_max;
+	iovec[1].iov_base = tag_str;
+	iovec[1].iov_len  = tag_max;
+	iovec[2].iov_base = pid_sep1;
+	iovec[2].iov_len  = pid_sep1_max;
+	iovec[3].iov_base = pid_str;
+	iovec[3].iov_len  = pid_max;
+	iovec[4].iov_base = pid_sep2;
+	iovec[4].iov_len  = pid_sep2_max;
+	iovec[5].iov_base = sd;
+	iovec[5].iov_len  = sd_max;
+	iovec[6].iov_base = dataptr;
+	iovec[6].iov_len  = max;
+	iovec[7].iov_base = "\n"; /* insert a \n at the end of the message */
+	iovec[7].iov_len  = 1;
+
+	if (logsrv->addr.ss_family == AF_UNSPEC) {
+		/* the target is a direct file descriptor. While writev() guarantees
+		 * to write everything, it doesn't guarantee that it will not be
+		 * interrupted while doing so. This occasionally results in interleaved
+		 * messages when the output is a tty, hence the lock. There's no real
+		 * performance concern here for such type of output.
+		 */
+		HA_SPIN_LOCK(LOGSRV_LOCK, &logsrv->lock);
+		sent = writev(*plogfd, iovec, 8);
+		HA_SPIN_UNLOCK(LOGSRV_LOCK, &logsrv->lock);
+	}
+	else {
+		msghdr.msg_name = (struct sockaddr *)&logsrv->addr;
+		msghdr.msg_namelen = get_addr_len(&logsrv->addr);
+
+		sent = sendmsg(*plogfd, &msghdr, MSG_DONTWAIT | MSG_NOSIGNAL);
+	}
+
+	if (sent < 0) {
+		static char once;
+
+		if (errno == EAGAIN)
+			_HA_ATOMIC_ADD(&dropped_logs, 1);
+		else if (!once) {
+			once = 1; /* note: no need for atomic ops here */
+			ha_alert("sendmsg()/writev() failed in logger #%d: %s (errno=%d)\n",
+					 nblogger, strerror(errno), errno);
+		}
+	}
+}
+
+/*
+ * This function sends a syslog message.
+ * It doesn't care about errors nor does it report them.
+ * The arguments <sd> and <sd_size> are used for the structured-data part
+ * in RFC5424 formatted syslog messages.
+ */
+void __send_log(struct proxy *p, int level, char *message, size_t size, char *sd, size_t sd_size)
+{
+	struct list *logsrvs = NULL;
+	struct logsrv *logsrv;
+	int nblogger;
+	static THREAD_LOCAL int curr_pid;
+	static THREAD_LOCAL char pidstr[100];
+	static THREAD_LOCAL struct buffer pid;
+	struct buffer *tag = &global.log_tag;
+
 	if (p == NULL) {
 		if (!LIST_ISEMPTY(&global.logsrvs)) {
 			logsrvs = &global.logsrvs;
@@ -1366,226 +1742,36 @@ void __send_log(struct proxy *p, int level, char *message, size_t size, char *sd
 
 	/* Send log messages to syslog server. */
 	nblogger = 0;
-	list_for_each_entry(tmp, logsrvs, list) {
-		const struct logsrv *logsrv = tmp;
-		int *plogfd;
-		char *pid_sep1 = "", *pid_sep2 = "";
-		char logheader_short[3];
-		int sent;
-		int maxlen;
-		int hdr_max = 0;
-		int tag_max = 0;
-		int pid_sep1_max = 0;
-		int pid_max = 0;
-		int pid_sep2_max = 0;
-		int sd_max = 0;
-		int max = 0;
-
-		nblogger++;
+	list_for_each_entry(logsrv, logsrvs, list) {
+		static THREAD_LOCAL int in_range = 1;
 
 		/* we can filter the level of the messages that are sent to each logger */
 		if (level > logsrv->level)
 			continue;
 
-		if (logsrv->addr.ss_family == AF_UNSPEC) {
-			/* the socket's address is a file descriptor */
-			plogfd = (int *)&((struct sockaddr_in *)&logsrv->addr)->sin_addr.s_addr;
-			if (unlikely(!((struct sockaddr_in *)&logsrv->addr)->sin_port)) {
-				/* FD not yet initialized to non-blocking mode.
-				 * DON'T DO IT ON A TERMINAL!
-				 */
-				if (!isatty(*plogfd))
-					fcntl(*plogfd, F_SETFL, O_NONBLOCK);
-				((struct sockaddr_in *)&logsrv->addr)->sin_port = 1;
-			}
-		}
-		else if (logsrv->addr.ss_family == AF_UNIX)
-			plogfd = &logfdunix;
-		else
-			plogfd = &logfdinet;
+		if (logsrv->lb.smp_rgs) {
+			struct smp_log_range *curr_rg;
 
-		if (unlikely(*plogfd < 0)) {
-			/* socket not successfully initialized yet */
-			if ((*plogfd = socket(logsrv->addr.ss_family, SOCK_DGRAM,
-			                      (logsrv->addr.ss_family == AF_UNIX) ? 0 : IPPROTO_UDP)) < 0) {
-				static char once;
-
-				if (!once) {
-					once = 1; /* note: no need for atomic ops here */
-					ha_alert("socket() failed in logger #%d: %s (errno=%d)\n",
-					         nblogger, strerror(errno), errno);
+			HA_SPIN_LOCK(LOGSRV_LOCK, &logsrv->lock);
+			curr_rg = &logsrv->lb.smp_rgs[logsrv->lb.curr_rg];
+			in_range = in_smp_log_range(curr_rg, logsrv->lb.curr_idx);
+			if (in_range) {
+				/* Let's consume this range. */
+				curr_rg->curr_idx = (curr_rg->curr_idx + 1) % curr_rg->sz;
+				if (!curr_rg->curr_idx) {
+					/* If consumed, let's select the next range. */
+					logsrv->lb.curr_rg = (logsrv->lb.curr_rg + 1) % logsrv->lb.smp_rgs_sz;
 				}
-				continue;
-			} else {
-				/* we don't want to receive anything on this socket */
-				setsockopt(*plogfd, SOL_SOCKET, SO_RCVBUF, &zero, sizeof(zero));
-				/* does nothing under Linux, maybe needed for others */
-				shutdown(*plogfd, SHUT_RD);
-				fcntl(*plogfd, F_SETFD, fcntl(*plogfd, F_GETFD, FD_CLOEXEC) | FD_CLOEXEC);
-			}
-		}
-
-		switch (logsrv->format) {
-		case LOG_FORMAT_RFC3164:
-			hdr = logheader;
-			hdr_ptr = update_log_hdr(time);
-			break;
-
-		case LOG_FORMAT_RFC5424:
-			hdr = logheader_rfc5424;
-			hdr_ptr = update_log_hdr_rfc5424(time);
-			sd_max = sd_size; /* the SD part allowed only in RFC5424 */
-			break;
-
-		case LOG_FORMAT_SHORT:
-			/* all fields are known, skip the header generation */
-			hdr = logheader_short;
-			hdr[0] = '<';
-			hdr[1] = '0' + MAX(level, logsrv->minlvl);
-			hdr[2] = '>';
-			hdr_ptr = hdr;
-			hdr_max = 3;
-			maxlen = logsrv->maxlen - hdr_max;
-			max = MIN(size, maxlen) - 1;
-			goto send;
-
-		case LOG_FORMAT_RAW:
-			/* all fields are known, skip the header generation */
-			hdr_ptr = hdr = "";
-			hdr_max = 0;
-			maxlen = logsrv->maxlen;
-			max = MIN(size, maxlen) - 1;
-			goto send;
-
-		default:
-			continue; /* must never happen */
-		}
-
-		hdr_size = hdr_ptr - hdr;
-
-		/* For each target, we may have a different facility.
-		 * We can also have a different log level for each message.
-		 * This induces variations in the message header length.
-		 * Since we don't want to recompute it each time, nor copy it every
-		 * time, we only change the facility in the pre-computed header,
-		 * and we change the pointer to the header accordingly.
-		 */
-		fac_level = (logsrv->facility << 3) + MAX(level, logsrv->minlvl);
-		hdr_ptr = hdr + 3; /* last digit of the log level */
-		do {
-			*hdr_ptr = '0' + fac_level % 10;
-			fac_level /= 10;
-			hdr_ptr--;
-		} while (fac_level && hdr_ptr > hdr);
-		*hdr_ptr = '<';
-
-		hdr_max = hdr_size - (hdr_ptr - hdr);
-
-		/* time-based header */
-		if (unlikely(hdr_size >= logsrv->maxlen)) {
-			hdr_max = MIN(hdr_max, logsrv->maxlen) - 1;
-			sd_max = 0;
-			goto send;
-		}
-
-		maxlen = logsrv->maxlen - hdr_max;
-
-		/* tag */
-		tag_max = tag->data;
-		if (unlikely(tag_max >= maxlen)) {
-			tag_max = maxlen - 1;
-			sd_max = 0;
-			goto send;
-		}
-
-		maxlen -= tag_max;
-
-		/* first pid separator */
-		pid_sep1_max = log_formats[logsrv->format].pid.sep1.data;
-		if (unlikely(pid_sep1_max >= maxlen)) {
-			pid_sep1_max = maxlen - 1;
-			sd_max = 0;
-			goto send;
-		}
-
-		pid_sep1 = log_formats[logsrv->format].pid.sep1.area;
-		maxlen -= pid_sep1_max;
-
-		/* pid */
-		pid_max = pid.data;
-		if (unlikely(pid_max >= maxlen)) {
-			pid_max = maxlen - 1;
-			sd_max = 0;
-			goto send;
-		}
-
-		maxlen -= pid_max;
-
-		/* second pid separator */
-		pid_sep2_max = log_formats[logsrv->format].pid.sep2.data;
-		if (unlikely(pid_sep2_max >= maxlen)) {
-			pid_sep2_max = maxlen - 1;
-			sd_max = 0;
-			goto send;
-		}
-
-		pid_sep2 = log_formats[logsrv->format].pid.sep2.area;
-		maxlen -= pid_sep2_max;
-
-		/* structured-data */
-		if (sd_max >= maxlen) {
-			sd_max = maxlen - 1;
-			goto send;
-		}
-
-		max = MIN(size, maxlen - sd_max) - 1;
-send:
-		iovec[0].iov_base = hdr_ptr;
-		iovec[0].iov_len  = hdr_max;
-		iovec[1].iov_base = tag->area;
-		iovec[1].iov_len  = tag_max;
-		iovec[2].iov_base = pid_sep1;
-		iovec[2].iov_len  = pid_sep1_max;
-		iovec[3].iov_base = pid.area;
-		iovec[3].iov_len  = pid_max;
-		iovec[4].iov_base = pid_sep2;
-		iovec[4].iov_len  = pid_sep2_max;
-		iovec[5].iov_base = sd;
-		iovec[5].iov_len  = sd_max;
-		iovec[6].iov_base = dataptr;
-		iovec[6].iov_len  = max;
-		iovec[7].iov_base = "\n"; /* insert a \n at the end of the message */
-		iovec[7].iov_len  = 1;
-
-		if (logsrv->addr.ss_family == AF_UNSPEC) {
-			/* the target is a direct file descriptor */
-			sent = writev(*plogfd, iovec, 8);
-		}
-		else {
-			msghdr.msg_name = (struct sockaddr *)&logsrv->addr;
-			msghdr.msg_namelen = get_addr_len(&logsrv->addr);
-
-			sent = sendmsg(*plogfd, &msghdr, MSG_DONTWAIT | MSG_NOSIGNAL);
-		}
-
-		if (sent < 0) {
-			static char once;
-
-			if (errno == EAGAIN)
-				HA_ATOMIC_ADD(&dropped_logs, 1);
-			else if (!once) {
-				once = 1; /* note: no need for atomic ops here */
-				ha_alert("sendmsg()/writev() failed in logger #%d: %s (errno=%d)\n",
-				         nblogger, strerror(errno), errno);
 			}
+			logsrv->lb.curr_idx = (logsrv->lb.curr_idx + 1) % logsrv->lb.smp_sz;
+			HA_SPIN_UNLOCK(LOGSRV_LOCK, &logsrv->lock);
 		}
+		if (in_range)
+			__do_send_log(logsrv, ++nblogger, pid.area, pid.data, level,
+			              message, size, sd, sd_size, tag->area, tag->data);
 	}
 }
 
-extern fd_set hdr_encode_map[];
-extern fd_set url_encode_map[];
-extern fd_set http_encode_map[];
-
 
 const char sess_cookie[8]     = "NIDVEOU7";	/* No cookie, Invalid cookie, cookie for a Down server, Valid cookie, Expired cookie, Old cookie, Unused, unknown */
 const char sess_set_cookie[8] = "NPDIRU67";	/* No set-cookie, Set-cookie found and left unchanged (passive),
@@ -1619,7 +1805,7 @@ static void init_log()
 
 	tmp = "\"\\]";
 	while (*tmp) {
-		FD_SET(*tmp, rfc5424_escape_map);
+		ha_bit_set(*tmp, rfc5424_escape_map);
 		tmp++;
 	}
 
@@ -1631,23 +1817,23 @@ static void init_log()
 	memset(hdr_encode_map, 0, sizeof(hdr_encode_map));
 	memset(url_encode_map, 0, sizeof(url_encode_map));
 	for (i = 0; i < 32; i++) {
-		FD_SET(i, hdr_encode_map);
-		FD_SET(i, url_encode_map);
+		ha_bit_set(i, hdr_encode_map);
+		ha_bit_set(i, url_encode_map);
 	}
 	for (i = 127; i < 256; i++) {
-		FD_SET(i, hdr_encode_map);
-		FD_SET(i, url_encode_map);
+		ha_bit_set(i, hdr_encode_map);
+		ha_bit_set(i, url_encode_map);
 	}
 
 	tmp = "\"#{|}";
 	while (*tmp) {
-		FD_SET(*tmp, hdr_encode_map);
+		ha_bit_set(*tmp, hdr_encode_map);
 		tmp++;
 	}
 
 	tmp = "\"#";
 	while (*tmp) {
-		FD_SET(*tmp, url_encode_map);
+		ha_bit_set(*tmp, url_encode_map);
 		tmp++;
 	}
 
@@ -1672,29 +1858,21 @@ static void init_log()
 	 */
 	memset(http_encode_map, 0, sizeof(http_encode_map));
 	for (i = 0x00; i <= 0x08; i++)
-		FD_SET(i, http_encode_map);
+		ha_bit_set(i, http_encode_map);
 	for (i = 0x0a; i <= 0x1f; i++)
-		FD_SET(i, http_encode_map);
-	FD_SET(0x7f, http_encode_map);
+		ha_bit_set(i, http_encode_map);
+	ha_bit_set(0x7f, http_encode_map);
 }
 
 INITCALL0(STG_PREPARE, init_log);
 
-static int init_log_buffers_per_thread()
-{
-	return init_log_buffers();
-}
-
-static void deinit_log_buffers_per_thread()
-{
-	deinit_log_buffers();
-}
-
 /* Initialize log buffers used for syslog messages */
 int init_log_buffers()
 {
 	logheader = my_realloc2(logheader, global.max_syslog_len + 1);
+	logheader_end = NULL;
 	logheader_rfc5424 = my_realloc2(logheader_rfc5424, global.max_syslog_len + 1);
+	logheader_rfc5424_end = NULL;
 	logline = my_realloc2(logline, global.max_syslog_len + 1);
 	logline_rfc5424 = my_realloc2(logline_rfc5424, global.max_syslog_len + 1);
 	if (!logheader || !logline_rfc5424 || !logline || !logline_rfc5424)
@@ -1711,7 +1889,7 @@ void deinit_log_buffers()
 	free(logheader_rfc5424);
 	free(logline);
 	free(logline_rfc5424);
-	tmp_startup_logs = HA_ATOMIC_XCHG(&startup_logs, NULL);
+	tmp_startup_logs = _HA_ATOMIC_XCHG(&startup_logs, NULL);
 	free(tmp_startup_logs);
 
 	logheader         = NULL;
@@ -1774,7 +1952,7 @@ int sess_build_logline(struct session *sess, struct stream *s, char *dst, size_t
 		txn = NULL;
 		be_conn = NULL;
 		s_flags = SF_ERR_PRXCOND | SF_FINST_R;
-		uniq_id = HA_ATOMIC_XADD(&global.req_count, 1);
+		uniq_id = _HA_ATOMIC_XADD(&global.req_count, 1);
 
 		/* prepare a valid log structure */
 		tmp_strm_log.tv_accept = sess->tv_accept;
@@ -2019,7 +2197,6 @@ int sess_build_logline(struct session *sess, struct stream *s, char *dst, size_t
 				break;
 
 			case LOG_FMT_TS: // %Ts
-				get_gmtime(logs->accept_date.tv_sec, &tm);
 				if (tmp->options & LOG_OPT_HEXA) {
 					iret = snprintf(tmplog, dst + maxsize - tmplog, "%04X", (unsigned int)logs->accept_date.tv_sec);
 					if (iret < 0 || iret > dst + maxsize - tmplog)
@@ -2772,7 +2949,7 @@ void strm_log(struct stream *s)
 
 	size = build_logline(s, logline, global.max_syslog_len, &sess->fe->logformat);
 	if (size > 0) {
-		HA_ATOMIC_ADD(&sess->fe->log_count, 1);
+		_HA_ATOMIC_ADD(&sess->fe->log_count, 1);
 		__send_log(sess->fe, level, logline, size + 1, logline_rfc5424, sd_size);
 		s->logs.logwait = 0;
 	}
@@ -2810,7 +2987,7 @@ void sess_log(struct session *sess)
 
 	size = sess_build_logline(sess, NULL, logline, global.max_syslog_len, &sess->fe->logformat);
 	if (size > 0) {
-		HA_ATOMIC_ADD(&sess->fe->log_count, 1);
+		_HA_ATOMIC_ADD(&sess->fe->log_count, 1);
 		__send_log(sess->fe, level, logline, size + 1, logline_rfc5424, sd_size);
 	}
 }
@@ -2837,8 +3014,8 @@ static struct cli_kw_list cli_kws = {{ },{
 
 INITCALL1(STG_REGISTER, cli_register_kw, &cli_kws);
 
-REGISTER_PER_THREAD_INIT(init_log_buffers_per_thread);
-REGISTER_PER_THREAD_DEINIT(deinit_log_buffers_per_thread);
+REGISTER_PER_THREAD_ALLOC(init_log_buffers);
+REGISTER_PER_THREAD_FREE(deinit_log_buffers);
 
 /*
  * Local variables:
