@@ -306,10 +306,20 @@ static struct h1s *h1s_create(struct h1c *h1c, struct conn_stream *cs, struct se
 		if (!sess)
 			sess = h1c->conn->owner;
 
-		h1s->csinfo.create_date = sess->accept_date;
-		h1s->csinfo.tv_create   = sess->tv_accept;
-		h1s->csinfo.t_handshake = sess->t_handshake;
-		h1s->csinfo.t_idle      = -1;
+		/* Timers for subsequent sessions on the same HTTP 1.x connection
+		 * measure from `now`, not from the connection accept time */
+		if (h1s->flags & H1S_F_NOT_FIRST) {
+			h1s->csinfo.create_date = date;
+			h1s->csinfo.tv_create   = now;
+			h1s->csinfo.t_handshake = 0;
+			h1s->csinfo.t_idle      = -1;
+		}
+		else {
+			h1s->csinfo.create_date = sess->accept_date;
+			h1s->csinfo.tv_create   = sess->tv_accept;
+			h1s->csinfo.t_handshake = sess->t_handshake;
+			h1s->csinfo.t_idle      = -1;
+		}
 	}
 	else {
 		if (h1c->px->options2 & PR_O2_RSPBUG_OK)
@@ -465,6 +475,10 @@ static void h1_release(struct h1c *h1c)
 
 		if (conn && h1c->flags & H1C_F_UPG_H2C) {
 			h1c->flags &= ~H1C_F_UPG_H2C;
+			/* Make sure we're no longer subscribed to anything */
+			if (h1c->wait_event.events)
+				conn->xprt->unsubscribe(conn, conn->xprt_ctx,
+				    h1c->wait_event.events, &h1c->wait_event);
 			if (conn_upgrade_mux_fe(conn, NULL, &h1c->ibuf, ist("h2"), PROTO_MODE_HTX) != -1) {
 				/* connection successfully upgraded to H2, this
 				 * mux was already released */
@@ -1386,6 +1400,9 @@ static size_t h1_process_input(struct h1c *h1c, struct buffer *buf, size_t count
 	}
 
 	data = htx->data;
+	if (h1s->flags & errflag)
+		goto end;
+
 	do {
 		size_t used = htx_used_space(htx);
 
@@ -1503,6 +1520,9 @@ static size_t h1_process_output(struct h1c *h1c, struct buffer *buf, size_t coun
 		errflag = H1S_F_REQ_ERROR;
 	}
 
+	if (h1s->flags & errflag)
+		goto end;
+
 	/* the htx is non-empty thus has at least one block */
 	blk = htx_get_head_blk(chn_htx);
 
@@ -1682,11 +1702,14 @@ static size_t h1_process_output(struct h1c *h1c, struct buffer *buf, size_t coun
 					h1m->flags |= (H1_MF_NO_PHDR|H1_MF_CLEAN_CONN_HDR);
 					h1s->flags &= ~H1S_F_HAVE_O_CONN;
 				}
+				else if ((h1m->flags & H1_MF_RESP) &&  h1s->meth == HTTP_METH_HEAD)
+					h1m->state = H1_MSG_DONE;
 				else
 					h1m->state = H1_MSG_DATA;
 				break;
 
 			case H1_MSG_DATA:
+			case H1_MSG_TUNNEL:
 				if (type == HTX_BLK_EOM) {
 					/* Chunked message without explicit trailers */
 					if (h1m->flags & H1_MF_CHNK) {
@@ -1696,7 +1719,9 @@ static size_t h1_process_output(struct h1c *h1c, struct buffer *buf, size_t coun
 					goto done;
 				}
 				else if (type == HTX_BLK_EOT || type == HTX_BLK_TLR) {
-					if (!chunk_memcat(&tmp, "0\r\n", 3))
+					/* If the message is not chunked, never
+					 * add the last chunk. */
+					if ((h1m->flags & H1_MF_CHNK) && !chunk_memcat(&tmp, "0\r\n", 3))
 						goto copy;
 					goto trailers;
 				}
@@ -1715,6 +1740,11 @@ static size_t h1_process_output(struct h1c *h1c, struct buffer *buf, size_t coun
 					goto error;
 			  trailers:
 				h1m->state = H1_MSG_TRAILERS;
+				/* If the message is not chunked, ignore
+				 * trailers. It may happen with H2 messages. */
+				if (!(h1m->flags & H1_MF_CHNK))
+					break;
+
 				if (type == HTX_BLK_EOT) {
 					if (!chunk_memcat(&tmp, "\r\n", 2))
 						goto copy;
@@ -1808,10 +1838,8 @@ static int h1_recv(struct h1c *h1c)
 	if (h1c->wait_event.events & SUB_RETRY_RECV)
 		return (b_data(&h1c->ibuf));
 
-	if (!(conn->flags & CO_FL_ERROR) && h1c->flags & H1C_F_CS_WAIT_CONN) {
-		conn->xprt->subscribe(conn, conn->xprt_ctx, SUB_RETRY_RECV, &h1c->wait_event);
+	if (!(conn->flags & CO_FL_ERROR) && h1c->flags & H1C_F_CS_WAIT_CONN)
 		return 0;
-	}
 
 	if (!h1_recv_allowed(h1c)) {
 		rcvd = 1;
@@ -1958,7 +1986,7 @@ static int h1_process(struct h1c * h1c)
 	}
 
 	if (!h1s) {
-		if (h1c->flags & H1C_F_CS_ERROR   ||
+		if (h1c->flags & (H1C_F_CS_ERROR|H1C_F_CS_SHUTDOWN) ||
 		    conn->flags & (CO_FL_ERROR | CO_FL_SOCK_WR_SH) ||
 		    conn_xprt_read0_pending(conn))
 			goto release;
@@ -2147,6 +2175,18 @@ static void h1_detach(struct conn_stream *cs)
 
 	if (conn_is_back(h1c->conn) && has_keepalive &&
 	    !(h1c->conn->flags & (CO_FL_ERROR | CO_FL_SOCK_RD_SH | CO_FL_SOCK_WR_SH))) {
+		/* If there are any excess server data in the input buffer,
+		 * release it and close the connection ASAP (some data may
+		 * remain in the output buffer). This happens if a server sends
+		 * invalid responses. So in such case, we don't want to reuse
+		 * the connection
+		 */
+		if (b_data(&h1c->ibuf)) {
+			h1_release_buf(h1c, &h1c->ibuf);
+			h1c->flags |= H1C_F_CS_SHUTW_NOW;
+			goto release;
+		}
+
 		/* Never ever allow to reuse a connection from a non-reuse backend */
 		if ((h1c->px->options & PR_O_REUSE_MASK) == PR_O_REUSE_NEVR)
 			h1c->conn->flags |= CO_FL_PRIVATE;
@@ -2192,9 +2232,12 @@ static void h1_detach(struct conn_stream *cs)
 		}
 	}
 
-	/* We don't want to close right now unless the connection is in error */
+  release:
+	/* We don't want to close right now unless the connection is in error or shut down for writes */
 	if ((h1c->flags & (H1C_F_CS_ERROR|H1C_F_CS_SHUTDOWN|H1C_F_UPG_H2C)) ||
-	    (h1c->conn->flags & CO_FL_ERROR) || !h1c->conn->owner)
+	    (h1c->conn->flags & (CO_FL_ERROR|CO_FL_SOCK_WR_SH)) ||
+	    ((h1c->flags & H1C_F_CS_SHUTW_NOW) && !b_data(&h1c->obuf)) ||
+	    !h1c->conn->owner)
 		h1_release(h1c);
 	else {
 		tasklet_wakeup(h1c->wait_event.tasklet);
