@@ -482,10 +482,6 @@ static void stream_int_notify(struct stream_interface *si)
 				ic->rex = tick_add_ifset(now_ms, ic->rto);
 	}
 
-	if ((sio->flags & SI_FL_RXBLK_ROOM) &&
-	    ((oc->flags & CF_WRITE_PARTIAL) || channel_is_empty(oc)))
-		si_rx_room_rdy(sio);
-
 	if (oc->flags & CF_DONT_READ)
 		si_rx_chan_blk(sio);
 	else
@@ -562,6 +558,16 @@ static void stream_int_notify(struct stream_interface *si)
 		task->expire = tick_first((tick_is_expired(task->expire, now_ms) ? 0 : task->expire),
 					  tick_first(tick_first(ic->rex, ic->wex),
 						     tick_first(oc->rex, oc->wex)));
+
+		task->expire = tick_first(task->expire, ic->analyse_exp);
+		task->expire = tick_first(task->expire, oc->analyse_exp);
+
+		if (si->exp)
+			task->expire = tick_first(task->expire, si->exp);
+
+		if (sio->exp)
+			task->expire = tick_first(task->expire, sio->exp);
+
 		task_queue(task);
 	}
 	if (ic->flags & CF_READ_ACTIVITY)
@@ -669,23 +675,14 @@ int si_cs_send(struct conn_stream *cs)
 
 	if (oc->pipe && conn->xprt->snd_pipe && conn->mux->snd_pipe) {
 		ret = conn->mux->snd_pipe(cs, oc->pipe);
-		if (ret > 0) {
-			oc->flags |= CF_WRITE_PARTIAL | CF_WROTE_DATA;
+		if (ret > 0)
 			did_send = 1;
-			if (si->state == SI_ST_CON)
-				si->state = SI_ST_RDY;
-		}
 
 		if (!oc->pipe->data) {
 			put_pipe(oc->pipe);
 			oc->pipe = NULL;
 		}
 
-		if (conn->flags & CO_FL_ERROR || cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING)) {
-			si->flags |= SI_FL_ERR;
-			return 1;
-		}
-
 		if (oc->pipe)
 			goto end;
 	}
@@ -752,10 +749,6 @@ int si_cs_send(struct conn_stream *cs)
 		ret = cs->conn->mux->snd_buf(cs, &oc->buf, co_data(oc), send_flag);
 		if (ret > 0) {
 			did_send = 1;
-			oc->flags |= CF_WRITE_PARTIAL | CF_WROTE_DATA;
-			if (si->state == SI_ST_CON)
-				si->state = SI_ST_RDY;
-
 			co_set_data(oc, co_data(oc) - ret);
 			c_realign_if_empty(oc);
 
@@ -763,19 +756,26 @@ int si_cs_send(struct conn_stream *cs)
 				/* Always clear both flags once everything has been sent, they're one-shot */
 				oc->flags &= ~(CF_EXPECT_MORE | CF_SEND_DONTWAIT);
 			}
-
 			/* if some data remain in the buffer, it's only because the
 			 * system buffers are full, we will try next time.
 			 */
 		}
-
-		if (conn->flags & CO_FL_ERROR || cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING)) {
-			si->flags |= SI_FL_ERR;
-			return 1;
-		}
 	}
 
  end:
+	if (did_send) {
+		oc->flags |= CF_WRITE_PARTIAL | CF_WROTE_DATA;
+		if (si->state == SI_ST_CON)
+			si->state = SI_ST_RDY;
+
+		si_rx_room_rdy(si_opposite(si));
+	}
+
+	if (conn->flags & CO_FL_ERROR || cs->flags & (CS_FL_ERROR|CS_FL_ERR_PENDING)) {
+		si->flags |= SI_FL_ERR;
+		return 1;
+	}
+
 	/* We couldn't send all of our data, let the mux know we'd like to send more */
 	if (!channel_is_empty(oc))
 		conn->mux->subscribe(cs, SUB_RETRY_SEND, &si->wait_event);
@@ -928,8 +928,7 @@ void si_sync_send(struct stream_interface *si)
 	if (cs->conn->flags & CO_FL_ERROR)
 		return;
 
-	if (si_cs_send(cs))
-		si_rx_room_rdy(si_opposite(si));
+	si_cs_send(cs);
 }
 
 /* Updates at once the channel flags, and timers of both stream interfaces of a
